# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Define utility classes and functions for cloudsync package."""

import os
import StringIO


def to_utf8(string):
    """Convert "string" to UTF-8 string if it "Unicode"."""

    if isinstance(string, unicode):
        return string.encode("UTF-8")
    return string


def get_stream(str_data):
    """Return the stream instance which include "str_data"."""

    data_stream = StringIO.StringIO()
    data_stream.write(str_data)
    data_stream.seek(0)
    return data_stream


class DataStream(object):
    """Class for accessing data as stream."""

    def _get_length(self, data, length):
        """Return a length of "data"."""

        if length >= 0:
            return length

        try:
            return len(data)
        except:
            pass

        try:
            length = 0
            if hasattr(data, "seek") and hasattr(data, "tell"):
                data.seek(0, os.SEEK_END)
                length = data.tell()
                data.seek(0)

            return length
        except:
            pass

        return 0

    def _get_data_stream(self, data):
        """Return the stream instance to access to "data"."""

        if hasattr(data, "read"):
            return data

        return get_stream(data)

    def __init__(self, data, length=-1):
        """Constructor."""

        self.data_stream = self._get_data_stream(data)
        self._length = self._get_length(data, length)

    def __len__(self):
        """Return a length of the stream."""

        return self._length

    def length(self):
        """Return a length of the stream."""

        return self._length

    def read(self, size=-1):
        """Return the data."""

        return self.data_stream.read(size)

    def seek(self, offset, whence=os.SEEK_SET):
        return self.data_stream.seek(offset, whence)


class UploadFileData(object):
    """Class for upload file data by "multipart/form-data"."""

    LINE_END = "\r\n"
    BOUNDARY_PREFIX = "--"
    BOUNDARY_SUFFIX = "--"

    FILE_NAME = "file"

    def _get_form_data_string(self, name, data=None, type=None, file_name=""):
        """Return the prefix string data for uploading."""

        form_data = self.BOUNDARY_PREFIX + self.boundary + self.LINE_END
        form_data += 'Content-Disposition: form-data; name="%s"' % (name,)
        if file_name:
            form_data += '; filename="%s"' % (file_name,)
        form_data += self.LINE_END

        if type != None:
            form_data += "Content-Type: %s%s" % (type, self.LINE_END)

        if data != None:
            form_data += self.LINE_END + data + self.LINE_END
        return form_data

    def _get_prefix_string(self, file_name):
        """Return the prefix string data for uploading."""

        prefix = self._get_form_data_string(self.FILE_NAME,
                                type="application/octet-stream",
                                file_name=file_name)
        prefix += self.LINE_END
        return prefix

    def _get_suffix_string(self):
        """Return the suffix string data for uploading."""

        return self.LINE_END + self.BOUNDARY_PREFIX + self.boundary + \
                        self.BOUNDARY_SUFFIX + self.LINE_END

    def _read(self, stream, size=-1):
        """Return the data in "stream"."""

        data = ""
        read_size = size
        while True:
            buf = stream.read(read_size)
            if len(buf) <= 0:
                return data
            data += buf
            if size > 0:
                read_size = size - len(buf)

    def __init__(self, file_name, data, length, boundary):
        """Constructor."""

        self.boundary = boundary
        self.headers = {"Content-Type":
                            "multipart/form-data; boundary=%s" % (boundary,)}
        self.prefix = DataStream(self._get_prefix_string(file_name))
        self.main = DataStream(data, length)
        self.suffix = DataStream(self._get_suffix_string())

        self.stream_list = (self.prefix, self.main, self.suffix)
        self.length = self.prefix.length() + self.main.length() + \
                        self.suffix.length()
        self.headers["Content-length"] = self.length
        self.reset_stream()

    def __len__(self):
        """Return a length of the upload data."""

        return self.length

    def get_headers(self):
        """Return dictionary for the request header."""

        return self.headers

    def reset_stream(self):
        """Reset the upload data stream."""

        self.prefix.seek(0)
        self.main.seek(0)
        self.suffix.seek(0)

    def read(self, size=-1):
        """Return the upload data."""

        data = ""
        read_size = size
        for stream in self.stream_list:
            data += self._read(stream, read_size)
            if size >= 0 and len(data) == size:
                return data

            if size > 0:
                read_size = size - len(data)

        return data
